/* 
 * mapfold.c --
 *
 *	This file contains the C-APIs for mapping and folding objects.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: mapfold.c,v 1.2 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"

/*
 *----------------------------------------------------------------------
 *
 * Listx_Map --
 *
 *	Scans the incoming list and transforms all objects in it. The
 *	transformation result is placed into another list, which is
 *	then returned as result of the call.
 *	
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	See above.
 *
 *----------------------------------------------------------------------
 */

int
Listx_Map (interp, listObj, mapProc, clientData, resultObj)
     Tcl_Interp*    interp;
     Tcl_Obj*       listObj;
     Listx_MapProc* mapProc;
     Tcl_Obj**      resultObj;
     ClientData     clientData;
{
    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    Tcl_Obj** listv;
    int       listc, result;
    long int  i;
    Tcl_Obj*  resMain, *resSub;


    result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);
    if (result != TCL_OK) {
	return result;
    }

    if (listc == 0) {
      /* The input is empty, and so is the output
       */

      resMain = Tcl_DuplicateObj (listObj);
      if (resMain == (Tcl_Obj*) NULL) {
	  return TCL_ERROR;
      } else {
	  *resultObj = resMain;
	  return TCL_OK;
      }
    }

    /*
     * Scan the list and apply the mapProc to each element.
     */

    resMain = Tcl_NewListObj (0, (Tcl_Obj**) NULL);
    if (resMain == (Tcl_Obj*) NULL) {
        return TCL_ERROR;
    }

    for (i = 0; i < listc; i++) {
        result = (*mapProc) (clientData, interp, listv [i], &resSub);
	if (result != TCL_OK) {
	    Tcl_DecrRefCount (resMain);
	    return result;
	}

        result = Tcl_ListObjAppendElement (interp, resMain, resSub);

	if (result != TCL_OK) {
	    Tcl_DecrRefCount (resMain);
	    return result;
	}
    }

    *resultObj = resMain;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Listx_Fold --
 *
 *	Scans the incoming list and accumulates/reduces all objects in it
 *	into a single object using a caller specified function.
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	See above.
 *
 *----------------------------------------------------------------------
 */

int
Listx_Fold (interp, listObj, initialObj, foldProc, clientData, resultObj)
     Tcl_Interp*     interp;
     Tcl_Obj*        listObj;
     Tcl_Obj*        initialObj;
     Listx_FoldProc* foldProc;
     Tcl_Obj**       resultObj;
     ClientData      clientData;
{
    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    Tcl_Obj** listv;
    int       listc, result;
    long int  i;
    Tcl_Obj*  accumulator, *res;


    result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);
    if (result != TCL_OK) {
	return result;
    }

    if (listc == 0) {
      /* The input is empty, and so is the output
       */

      res = Tcl_DuplicateObj (listObj);
      if (res == (Tcl_Obj*) NULL) {
	  return TCL_ERROR;
      } else {
	  *resultObj = res;
	  return TCL_OK;
      }
    }

    /*
     * Scan the list and apply the foldProc to each element
     * and the current accumulator.
     */

    accumulator = Tcl_DuplicateObj (initialObj);
    if (accumulator == (Tcl_Obj*) NULL) {
        return TCL_ERROR;
    }

    for (i = 0; i < listc; i++) {
	Tcl_IncrRefCount (accumulator);
        result = (*foldProc) (clientData, interp, accumulator, listv [i],
			      &res);
	Tcl_DecrRefCount (accumulator);
	if (result != TCL_OK) {
	    return result;
	}

	accumulator = res;
    }

    *resultObj = accumulator;
    return TCL_OK;
}

