/* 
 * search.c --
 *
 *	This file contains the C-API functionality for searching in lists.
 *
 * Copyright (c) 1987-1993 The Regents of the University of California.
 * Copyright (c) 1994-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: search.c,v 1.2 2000/01/14 23:44:16 aku Exp $
 */

#include "listxInt.h"

static int
Search _ANSI_ARGS_ ((Tcl_Interp* interp, Tcl_Obj* listObj, Tcl_Obj* patternObj,
		     long int offset, lSearchModes mode, int inverted,
		     int nocase, long int* idx));

/*
 *----------------------------------------------------------------------
 *
 * Listx_Search --
 *
 *	Searches the list object starting at offset for one or all items (not)
 *	matching the pattern. Employs either string comparison, glob
 *	expressions or regular expressions.
 *	
 *
 * Results:
 *	A standard tcl result code.
 *
 * Side effects:
 *	The found indices are stored in the provided list selector structure.
 *
 *----------------------------------------------------------------------
 */

int
Listx_Search (interp, listObj, patternObj, offset, mode, inverted, nocase,
	      all, ls)
     Tcl_Interp*   interp;
     Tcl_Obj*      listObj;
     Tcl_Obj*      patternObj;
     long int      offset;
     lSearchModes  mode;
     int           inverted;
     int           nocase;
     int           all;
     listSelector* ls;
{
  int      result;
  long int index;

  if (!all) {
    result = Search (interp, listObj, patternObj, offset, mode, inverted,
		     nocase, &index);

    if (result != TCL_OK) {
      return result;
    }

    if (index < 0) {
      ls->type                = LIST_SEL_MULTIPLE;
      ls->info.multiple.num   = 0;
      ls->info.multiple.index = (lIndex *) NULL;
    } else {
      ls->type                  = LIST_SEL_SINGLE;
      ls->info.index.offsetFrom = LOFF_START;
      ls->info.index.location   = index;
    }

    return result;
  } else {
    lIndex* idx;
    int  i, listc;
    Tcl_Obj** listv;

    result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);

    if (result != TCL_OK) {
      return result;
    }

    idx = (lIndex*) Tcl_Alloc (listc * sizeof (lIndex));
    i   = 0;

    while (offset < listc) {
      result = Search (interp, listObj, patternObj, offset, mode, inverted,
		       nocase, &index);

      if (result != TCL_OK) {
	Tcl_Free ((char*) idx);
	return result;
      }

      if (index < 0) {
	break;
      }

      idx [i].offsetFrom = LOFF_START;
      idx [i].location   = index;
      i++;
      offset = index + 1;
    }

    ls->type                = LIST_SEL_MULTIPLE;
    ls->info.multiple.num   = i;
    ls->info.multiple.index = idx;

  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Search --
 *
 *	Searches the list object starting at offset for an item (not)
 *	matching the pattern. Employs either string comparison, glob
 *	expressions or regular expressions.
 *	
 *
 * Results:
 *	A standard tcl result.
 *
 * Side effects:
 *	A list selector structure filled with the indices of the
 *	(not)matching elements.
 *
 *----------------------------------------------------------------------
 */

static int
Search (interp, listObj, patternObj, offset, mode, inverted, nocase, idx)
     Tcl_Interp*   interp;
     Tcl_Obj*      listObj;
     Tcl_Obj*      patternObj;
     long int      offset;
     lSearchModes  mode;
     int           inverted;
     int           nocase;
     long int*     idx;
{
    /*
     * Make sure the list argument is a list object and get its length and
     * a pointer to its array of element pointers.
     */

    Tcl_Obj** listv;
    int       listc, result, rflags, length, elemLen, match;
    long int  i, index;
    char*     bytes, *patternBytes;


    result = Tcl_ListObjGetElements(interp, listObj, &listc, &listv);
    if (result != TCL_OK) {
	return result;
    }

    index  = -1;
    rflags = TCL_REG_ADVANCED | TCL_REG_NOSUB | (nocase ? TCL_REG_NOCASE : 0);

    for (i = offset; i < listc; i++) {
	match = 0;

	switch (mode) {
	    case LSEARCH_EXACT: {
	        int minlen;

	        patternBytes = Tcl_GetStringFromObj (patternObj, &length);
		bytes        = Tcl_GetStringFromObj (listv[i],   &elemLen);

		/*
		 * This is the min length IN BYTES of the two strings
		 */
		minlen = (length < elemLen) ? length : elemLen;

		if (nocase) {
		    /*
		     * with -nocase we have to check true char length
		     * as it could be smaller than expected
		     */

		    length  = Tcl_NumUtfChars(patternBytes, length);
		    elemLen = Tcl_NumUtfChars(bytes,        elemLen);

		    minlen = (length < elemLen) ? length : elemLen;

		    match = Tcl_UtfNcasecmp(patternBytes, bytes,
					    (unsigned) minlen);
		} else {
		    match = memcmp(patternBytes, bytes, (unsigned) minlen);
		}

		if (match == 0) {
		    match = length - elemLen;
		}

		match = (match == 0 ? 1 : 0);
		break;
	    }
	    case LSEARCH_GLOB: {
	        patternBytes = Tcl_GetStringFromObj(patternObj, &length);
		match = Tcl_StringCaseMatch(Tcl_GetString(listv[i]),
					    patternBytes, nocase);
		break;
	    }
	    case LSEARCH_REGEXP: {
	        /* match = Tcl_RegExpMatchObj(interp, listv[i], patObj);
		 * Using the code below to allow us the usage of 'nocase'.
		 */

	        Tcl_RegExp re;

		re = Tcl_GetRegExpFromObj (interp, patternObj, rflags);
		if (re == NULL) {
		    return TCL_ERROR;
		}

		match = Tcl_RegExpExecObj (interp, re, listv [i],
					   0 /* offset */, 0 /* nmatches */,
					   0 /* flags */);
		if (match < 0) {
		    return TCL_ERROR;
		}
		break;
	    }
	}

	if (inverted) {
	    match = (match != 0 ? 0 : 1);
	}

	if (match != 0) {
	    index = i;
	    break;
	}
    }

    *idx = index;
    return TCL_OK;
}

