/* 
 * selectObj.c --
 *
 *	This file contains the implementation of the 'selector' Tcl_ObjType.
 *
 * Copyright (c) 1999 Andreas Kupries <a.kupries@westend.com>
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: setObj.c,v 1.2 1999/12/10 18:10:53 aku Exp $
 */

#include "listxInt.h"


/*
 * Prototypes for procedures defined later in this file:
 */

static void		DupSelectorInternalRep _ANSI_ARGS_ ((Tcl_Obj *srcPtr,
			    Tcl_Obj *copyPtr));
static void		FreeSelectorInternalRep _ANSI_ARGS_((Tcl_Obj *setPtr));
static int		SetSelectorFromAny _ANSI_ARGS_ ((Tcl_Interp *interp,
			    Tcl_Obj *objPtr));
static void		UpdateStringOfSelector _ANSI_ARGS_ ((Tcl_Obj *setPtr));

static void             CopyListSelector _ANSI_ARGS_ ((listSelector* CONST src,
						       listSelector* dst));

static int              ParseSlice _ANSI_ARGS_ ((Tcl_Interp* interp,
						 int listc, Tcl_Obj** listv,
						 listSelector* selRepPtr));

static int              GetIndex _ANSI_ARGS_ ((Tcl_Interp*   interp,
					       Tcl_Obj*      obj,
					       int* loc, int* fromEnd));

/*
 * The structure below defines the selector Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

Tcl_ObjType selectorObjType = {
    "selector",			/* name */
    FreeSelectorInternalRep,	/* freeIntRepProc */
    DupSelectorInternalRep,     /* dupIntRepProc */
    UpdateStringOfSelector,	/* updateStringProc */
    SetSelectorFromAny		/* setFromAnyProc */
};


/*
 * Our reference to the builtin list type is stored in 'ListxListType',
 * and initialized in init.c:Listx_Init.
 */

/* The value below is given to TclGetIntForIndex to denote 'end'. As that
 * function does not do range checking this allows me to move 'end'-number
 * to negative values with -1 == end, -2 == end - 1, etc.
 */

#define END_VALUE (-1)


/*
 *----------------------------------------------------------------------
 *
 * Listx_NewSelectorObj --
 *
 *	This procedure is normally called when not debugging: i.e., when
 *	TCL_MEM_DEBUG is not defined. It creates a new selector object from
 *	a listSelector structure.
 *
 *	When TCL_MEM_DEBUG is defined, this procedure just returns the
 *	result of calling the debugging version Listx_DbNewSelectorObj.
 *
 * Results:
 *	A new selector object is returned that is initialized from the
 *	listSelector structure given as argument. The new object's string
 *	representation is left NULL. The resulting new selector object has
 *	ref count 0.
 *
 * Side effects:
 *	None beyond memory allocation.
 *
 *----------------------------------------------------------------------
 */

#ifdef TCL_MEM_DEBUG
#undef Listx_NewSelectorObj

Tcl_Obj *
Listx_NewSelectorObj (ls)
     listSelector* CONST ls; /* The list selector information used to
			      * initialize the new object. */
{
    return Listx_DbNewSelectorObj (ls, "unknown", 0);
}

#else /* if not TCL_MEM_DEBUG */

Tcl_Obj *
Listx_NewSelectorObj (ls)
     listSelector* CONST ls; /* The list selector information used to
			      * initialize the new object. */
{
    register listSelector* selRepPtr;
    Tcl_Obj*               selectorObj;

    selectorObj = Tcl_NewObj ();
    Tcl_InvalidateStringRep(selectorObj);

    selRepPtr = (listSelector *) Tcl_Alloc (sizeof (listSelector));

    selectorObj->internalRep.otherValuePtr = (VOID *) selRepPtr;
    selectorObj->typePtr                   = &selectorObjType;

    CopyListSelector (ls, selRepPtr);

    return selectorObj;
}
#endif /* if TCL_MEM_DEBUG */

/*
 *----------------------------------------------------------------------
 *
 * Listx_DbNewSelectorObj --
 *
 *	This procedure is normally called when debugging: i.e., when
 *	TCL_MEM_DEBUG is defined. It creates new selector objects. It is the
 *	same as the Listx_NewSelectorObj procedure above except that it calls
 *	Tcl_DbCkalloc directly with the file name and line number from its
 *	caller. This simplifies debugging since then the checkmem command
 *	will report the correct file name and line number when reporting
 *	objects that haven't been freed.
 *
 *	When TCL_MEM_DEBUG is not defined, this procedure just returns the
 *	result of calling Listx_NewSelectorObj.
 *
 * Results:
 *	A new selector object is returned that is initialized from the
 *	listSelector structure given as argument. The new object's string
 *	representation is left NULL. The resulting new selector object has
 *	ref count 0.
 *
 * Side effects:
 *	None beyond memory allocation.
 *
 *----------------------------------------------------------------------
 */

#ifdef TCL_MEM_DEBUG

Tcl_Obj *
Listx_DbNewSelectorObj (ls, file, line)
     listSelector* CONST ls; /* The list selector information used to
			      * initialize the new object. */
     char *file;		/* The name of the source file calling this
				 * procedure; used for debugging. */
     int line;			/* Line number in the source file; used
				 * for debugging. */
{
    int           i, dummy;
    register listSelector* selRepPtr;
    Tcl_Obj*      selectorObj;

    selectorObj = Tcl_DbNewObj ();

    selRepPtr = (listSelector *) Tcl_Alloc (sizeof (listSelector));

    selectorObj->internalRep.otherValuePtr = (VOID *) selRepPtr;
    selectorObj->typePtr                   = &selectorObjType;

    CopyListSelector (ls, selRepPtr);

    return selectorObj;
}

#else /* if not TCL_MEM_DEBUG */

Tcl_Obj *
Listx_DbNewSelectorObj (ls, file, line)
     listSelector* CONST ls; /* The list selector information used to
			      * initialize the new object. */
     char *file;		/* The name of the source file calling this
				 * procedure; used for debugging. */
     int line;			/* Line number in the source file; used
				 * for debugging. */
{
    return Listx_NewSelectorObj (ls);
}
#endif /* TCL_MEM_DEBUG */


/*
 *----------------------------------------------------------------------
 *
 * Listx_GetSelectorFromObj --
 *
 *	Retrieve selector information from the object.
 *
 * Results:
 *	A standard tcl result code.
 *
 * Side effects:
 *	Sets the variable pointed to by 'ls' to the selector structure
 *	contained in the selector object.
 *
 *----------------------------------------------------------------------
 */

int
Listx_GetSelectorFromObj (interp, obj, ls)
     Tcl_Interp*    interp;
     Tcl_Obj*       obj;
     listSelector** ls;
{
  int result;

  result = Tcl_ConvertToType (interp, obj, &selectorObjType);

  if (result != TCL_OK) {
    return result;
  }

  *ls = (listSelector *) obj->internalRep.otherValuePtr;
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ListxConvertToFullIndex --
 *
 *	Deallocate the storage associated with a selector object's internal
 *	representation.
 *
 * Results:
 *	A standard tcl result code.
 *
 * Side effects:
 *	Sets 'fullIndex' to the absolute position in the list described
 *	by 'codedIndex' and 'endValue'. This happens *even if* there are
 *	errors. In case of ranges f.e. corrective action can be taken.
 *
 *----------------------------------------------------------------------
 */

int
ListxConvertToFullIndex (codedIndex, endValue, fullIndex)
     lIndex*     codedIndex;
     int         endValue;
     int*        fullIndex;
{
  int absIndex;

  if (codedIndex->offsetFrom == LOFF_END) {
    /* Convert end-relative position into absolute one.
     */

    absIndex = endValue-1 - codedIndex->location;
  } else {
    absIndex = codedIndex->location;
  }

  *fullIndex = absIndex;

  if ((absIndex < 0) || (absIndex >= endValue)) {
    /* Index out of range.
     */
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * FreeSelectorInternalRep --
 *
 *	Deallocate the storage associated with a selector object's internal
 *	representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees selPtr's listSelector* internal representation and sets selPtr's
 *	internalRep.otherValuePtr to NULL.
 *
 *----------------------------------------------------------------------
 */

static void
FreeSelectorInternalRep(selPtr)
    Tcl_Obj *selPtr;		/* Set object with internal rep to free. */
{
    register listSelector *selRepPtr =
      (listSelector *) selPtr->internalRep.otherValuePtr;

    if ((selRepPtr->type == LIST_SEL_MULTIPLE) &&
	(selRepPtr->info.multiple.num > 0)) {

      Tcl_Free((char *) selRepPtr->info.multiple.index);
    }

    Tcl_Free((char *) selRepPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * DupSelectorInternalRep --
 *
 *	Initialize the internal representation of a selector Tcl_Obj to a
 *	copy of the internal representation of an existing selector object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"srcPtr"s selector internal rep pointer should not be NULL and we
 *	assume it is not NULL. We set "copyPtr"s internal rep to a pointer
 *	to a newly allocated listSelector structure.
 *
 *----------------------------------------------------------------------
 */

static void
DupSelectorInternalRep(srcPtr, copyPtr)
    Tcl_Obj *srcPtr;		/* Object with internal rep to copy. */
    Tcl_Obj *copyPtr;		/* Object with internal rep to set. */
{
    register listSelector*  srcSelRepPtr = 
      (listSelector *) srcPtr->internalRep.otherValuePtr;

    register listSelector*  copySelRepPtr;

    /*
     * Allocate a new Selector structure and initialize it with a copy of the
     * source structure.
     */

    copySelRepPtr = (listSelector *) Tcl_Alloc (sizeof(listSelector));
    CopyListSelector (srcSelRepPtr, copySelRepPtr);
    
    copyPtr->internalRep.otherValuePtr = (VOID *) copySelRepPtr;
    copyPtr->typePtr = &selectorObjType;
}

/*
 *----------------------------------------------------------------------
 *
 * SetSelectorFromAny --
 *
 *	Attempt to generate a selector internal form for the Tcl object
 *	"objPtr".
 *
 * Results:
 *	The return value is TCL_OK or TCL_ERROR. If an error occurs during
 *	conversion, an error message is left in the interpreter's result
 *	unless "interp" is NULL.
 *
 * Side effects:
 *	If no error occurs, a selector is stored as "objPtr"s internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static int
SetSelectorFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    Tcl_Obj *objPtr;		/* The object to convert. */
{
    Tcl_ObjType*           oldTypePtr = objPtr->typePtr;
    int                    listc, result, i, fromEnd, idx, idxOff;
    register listSelector* selRepPtr = (listSelector*) NULL;
    Tcl_Obj**              listv;
    lIndex*                index = (lIndex *) NULL;

#define SLICE  "..."
#define STRIDE "|"

    /*
     * This is never called for an object which is already of type 'selector'.
     */

    /* Conversion from any object to a selector. First try to convert it
     * into a list and then from that into a selector. This will work for
     * 'single' selectors too, as they get converted into a single-item
     * list in this step. And as single selectors and single-item
     * multi-selectors are handled the same way this is no problem from
     * the semantical sidee as well.
     */

    result = Tcl_ConvertToType (interp, objPtr, ListxListType);

    if (result != TCL_OK) {
        return result;
    }

    result = Tcl_ListObjGetElements(interp, objPtr, &listc, &listv);

    if (result != TCL_OK) {
        return result;
    }

    selRepPtr = (listSelector *) Tcl_Alloc (sizeof (listSelector));

    if (listc == 0) {
      /* Empty list. Use an empty 'multiple' selector.
       */

      selRepPtr->type                = LIST_SEL_MULTIPLE;
      selRepPtr->info.multiple.num   = 0;
      selRepPtr->info.multiple.index = (lIndex*) NULL;

    } else if (listc == 1) {
      /* Single item list => Take the element and try to convert it either
       * as number or 'end'-number. A negative value represents 'end'-number.
       * Danger: -1 == end, -2 == end-1, etc. Or else the value 0 would be
       * ambiguous.
       */

      int idx;

      result = GetIndex (interp, listv [0], &idx, &fromEnd);

      if (result != TCL_OK) goto error;

      selRepPtr->type                  = LIST_SEL_SINGLE;
      selRepPtr->info.index.offsetFrom = (fromEnd ? LOFF_END : LOFF_START);
      selRepPtr->info.index.location   = idx;

    } else {
      /* More than one item. Check for the various possibilities to specify
       * slices first. Then try to convert everything as number or
       * 'end'-number.
       */

      result = ParseSlice (interp, listc, listv, selRepPtr);

      if (result != TCL_OK) {
	/* No slice, try to generate a multi-selector.
	 */

	index = (lIndex*) Tcl_Alloc (listc * sizeof (lIndex));

	for (i=0; i < listc; i++) {

	  result =
	    GetIndex (interp, listv [i], &idx, &idxOff);

	  if (result != TCL_OK) goto error;

	  index [i].offsetFrom = (idxOff ? LOFF_END : LOFF_START);
	  index [i].location   = idx;
	}

	selRepPtr->type                = LIST_SEL_MULTIPLE;
	selRepPtr->info.multiple.num   = listc;
	selRepPtr->info.multiple.index = index;
      }
    }

    /*
     * Free the old internalRep before setting the new one. We do this as
     * late as possible to allow the conversion code, in particular
     * Tcl_GetStringFromObj, to use that old internalRep.
     */

    if ((oldTypePtr != NULL) &&
	(oldTypePtr->freeIntRepProc != NULL)) {
      oldTypePtr->freeIntRepProc(objPtr);
    }

    objPtr->internalRep.otherValuePtr = (VOID *) selRepPtr;
    objPtr->typePtr                   = &selectorObjType;
    return TCL_OK;

error:
    if (selRepPtr != (listSelector*) NULL) {
      Tcl_Free ((char*) selRepPtr);
    }

    if (index != (lIndex*) NULL) {
      Tcl_Free ((char*) index);
    }

    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfSelector --
 *
 *	Update the string representation for a selector object.
 *	Note: This procedure does not invalidate an existing old string rep
 *	so storage will be lost if this has not already been done. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The object's string is set to a valid string that results from
 *	the selector-to-string conversion. This string will be empty if
 *	the selector has no elements. The selector internal representation
 *	should not be NULL and we assume it is not NULL.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfSelector (selPtr)
    Tcl_Obj *selPtr;	/* Selector object with string rep to update. */
{
  char buffer [TCL_INTEGER_SPACE+4];

  listSelector *selRepPtr =
    (listSelector *) selPtr->internalRep.otherValuePtr;

  register int i;
  char        *dst;
  int          len;

    /*
     * Pass 1:	Estimate space, gather flags.
     *		Dependent on the type of the selector.
     */

    selPtr->length = 1;

    switch (selRepPtr->type) {
    case LIST_SEL_SINGLE:
      selPtr->length = TCL_INTEGER_SPACE+4;
      break;

    case LIST_SEL_MULTIPLE:
      selPtr->length =
	((TCL_INTEGER_SPACE+4)+1) * selRepPtr->info.multiple.num;
      break;

    case LIST_SEL_RANGE:
      selPtr->length =
	(TCL_INTEGER_SPACE+4) * 3 + 4 /* ... | */ + 4 /* spaces */;
      break;
    }

    /*
     * Pass 2:	Copy into string rep buffer.
     *		Dependent on the type of the selector, again.
     */

    selPtr->bytes = Tcl_Alloc((unsigned) selPtr->length);
    dst           = selPtr->bytes;

    switch (selRepPtr->type) {
    case LIST_SEL_SINGLE:

      /* ... x < 0 => 'end'-x */

      len = TclFormatInt (buffer, selRepPtr->info.index);
      strcpy (dst, buffer); dst += len;
      *dst = ' ';
      dst++;
      break;

    case LIST_SEL_MULTIPLE:
      for (i=0; i < selRepPtr->info.multiple.num; i++) {
	len = TclFormatInt (buffer, selRepPtr->info.multiple.index [i]);
	strcpy (dst, buffer); dst += len;
	*dst = ' ';
	dst++;
      }
      break;

    case LIST_SEL_RANGE:
      len = TclFormatInt (buffer, selRepPtr->info.range.start);
      strcpy (dst, buffer); dst += len;
      *dst = ' ';
      dst++;

      strcpy (dst, SLICE); dst += 3;
      *dst = ' ';          dst++;

      len = TclFormatInt (buffer, selRepPtr->info.range.end);
      strcpy (dst, buffer); dst += len;
      *dst = ' ';
      dst++;

      strcpy (dst, STRIDE); dst ++;
      *dst = ' ';           dst ++;

      len = TclFormatInt (buffer, selRepPtr->info.range.stride);
      strcpy (dst, buffer); dst += len;
      *dst = ' ';
      dst++;
      break;
    }

    if (dst == selPtr->bytes) {
	*dst = 0;
    } else {
	dst --;
	*dst = 0;
    }

    selPtr->length = dst - selPtr->bytes;
}

/*
 *----------------------------------------------------------------------
 *
 * CopyListSelector --
 *
 *	Copies an existing list selector into another (uninitialized)
 *	structure. Note: Using an initialized selector structure as
 *	destination may cause leaking of memory.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The 'dst' structure is set to a copy of the 'src' information.
 *
 *----------------------------------------------------------------------
 */

static void
CopyListSelector (src, dst)
     listSelector* CONST src;
     listSelector*       dst;
{
    memcpy (dst, src, sizeof (listSelector));

    if ((dst->type == LIST_SEL_MULTIPLE) &&
	(dst->info.multiple.num > 0)) {

      /* Duplicate the array too.
       */

      dst->info.multiple.index =
	(lIndex*) Tcl_Alloc (dst->info.multiple.num * sizeof (lIndex));

      memcpy (dst->info.multiple.index,
	      src->info.multiple.index,
	      dst->info.multiple.num * sizeof (lIndex));
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ParseSlice --
 *
 *	Takes a list and searches it fors indicators showing that it
 *	contains a slice specification. If the hypothesis was found to
 *	be true the information from the specification is copied into
 *	the selector structure.
 *
 * Results:
 *	A standard tcl result code.
 *
 * Side effects:
 *	In case of success the selector structure is set to the relevant
 *	information from the slice specification.
 *
 *----------------------------------------------------------------------
 */

static int
ParseSlice (interp, listc, listv, selRepPtr)
     Tcl_Interp*   interp;
     int           listc;
     Tcl_Obj**     listv;
     listSelector* selRepPtr;
{
  int result, start, end, stride, startOff, endOff;

  /* Slices:
   *
   * from ... to | stride		5 items
   *
   * from ...    | stride		4 items
   *      ... to | stride		4 items
   *
   * from ...				2 items
   *      ... to			2 items
   *             | stride		2 items
   */

  if (listc == 5) {
    /*
     * from ... to | stride
     */

    result = GetIndex (interp, listv [0], &start, &startOff);
    if (result != TCL_OK) return TCL_ERROR;

    result = GetIndex (interp, listv [2], &end, &endOff);
    if (result != TCL_OK) return TCL_ERROR;

    result = Tcl_GetIntFromObj (interp, listv [4], &stride);
    if (result != TCL_OK) return TCL_ERROR;

    if (strcmp (SLICE,  Tcl_GetString (listv [1]))) return TCL_ERROR;
    if (strcmp (STRIDE, Tcl_GetString (listv [3]))) return TCL_ERROR;

  } else if (listc == 4) {
    /*
     *      from ... | stride
     *      ...  to  | stride
     */

    if (strcmp (STRIDE,   Tcl_GetString (listv [2]))) return TCL_ERROR;

    result = Tcl_GetIntFromObj (interp, listv [3], &stride);
    if (result != TCL_OK) return TCL_ERROR;

    if (0 == strcmp (SLICE, Tcl_GetString (listv [1]))) {

      result = GetIndex (interp, listv [0], &start, &startOff);
      if (result != TCL_OK) return TCL_ERROR;

      end    = 0;
      endOff = 1;

    } else if (0 == strcmp (SLICE, Tcl_GetString (listv [0]))) {

      result = GetIndex (interp, listv [1], &end, &endOff);
      if (result != TCL_OK) return TCL_ERROR;

      start    = 0;
      startOff = 0;

    } else {
      return TCL_ERROR;
    }

  } else if (listc == 2) {
    /*
     * from ...
     * ...  to
     * |    stride
     */

    if (0 == strcmp (SLICE, Tcl_GetString (listv [1]))) {
      result = GetIndex (interp, listv [0], &start, &startOff);
      if (result != TCL_OK) return TCL_ERROR;
      
      end    = 0;
      endOff = 1;
      stride = 1;

    } else if (0 == strcmp (SLICE, Tcl_GetString (listv [0]))) {
      result = GetIndex (interp, listv [1], &end, &endOff);
      if (result != TCL_OK) return TCL_ERROR;
      
      start    = 0;
      startOff = 0;
      stride   = 1;

    } else if (0 == strcmp (STRIDE, Tcl_GetString (listv [0]))) {

      result = Tcl_GetIntFromObj (interp, listv [1], &stride);
      if (result != TCL_OK) return TCL_ERROR;

      start    = 0;
      startOff = 0;
      end      = 0;
      endOff   = 1;

    } else {
      return TCL_ERROR;
    }
  } else {
    /* Can't be a slice, number of items is not correct */

    return TCL_ERROR;
  }


  if (stride == 0) {
    Tcl_AppendResult (interp, "Illegal stride 0", NULL);
    return TCL_ERROR;
  }

  selRepPtr->type                         = LIST_SEL_RANGE;
  selRepPtr->info.range.start.offsetFrom  = (startOff ? LOFF_END : LOFF_START);
  selRepPtr->info.range.start.location    = start;
  selRepPtr->info.range.end.offsetFrom    = (endOff   ? LOFF_END : LOFF_START);
  selRepPtr->info.range.end.location      = end;
  selRepPtr->info.range.stride = stride;

  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetIndex --
 *
 *	A helper procedure wrapping around TclGetIntForIndex.
 *	Required to reliably detect 'end-number'. TclGetIntForIndex handles
 *	them, but does not deliver information to the caller when it detects
 *	one. And we need that information so that a selector-object stores
 *	information not specific to a certain list, but usable with all lists.
 *
 * Results:
 *	A standard tcl result code.
 *
 * Side effects:
 *	In case of success 'loc' and 'fromEnd' are set to the extracted
 *	information.
 *
 *----------------------------------------------------------------------
 */

static int
GetIndex (interp, obj, loc, fromEnd)
     Tcl_Interp*   interp;
     Tcl_Obj*      obj;
     int*          loc;
     int*          fromEnd;
{
  /* Try simple 'Int' first. If that causes a problem defer to 'IntForIndex'.
   * If that returns something it has to be 'end-...'.
   */

  int res, index;

  res = Tcl_GetIntFromObj (interp, obj, &index);

  if (res == TCL_OK) {
    /* Normal index, positioned from the start.
     */

    *fromEnd = 0;
    *loc     = index;
    return TCL_OK;
  }

  Tcl_ResetResult (interp);
  res = TclGetIntForIndex (interp, obj, END_VALUE, &index);

  if (res == TCL_OK) {

    /* index:    -1 == end, -2 == end - 1, etc.
     * -index-1:  0 == end,  1 == end - 1, etc.
     */

    *fromEnd = 1;
    *loc     = -index-1;
    return TCL_OK;
  }

  return res;
}
